#include "CurrentMap.h"

// Constructors
CurrentMap::CurrentMap() { }
// Set up the memory for the tile maps
void CurrentMap::InitializeMemory() 
{
	// Set the registers for the first background to be tiled
	REG_BG0CNT = BG_64x64 | BG_COLOR_256 | BG_MAP_BASE(0) | BG_TILE_BASE(1);

	// Copy the tile sheet into memory
	u8* tileMemory = (u8*)BG_TILE_RAM(1);
	swiCopy(tilesheetPal, BG_PALETTE, tilesheetPalLen);
	swiCopy(tilesheetTiles, tileMemory, tilesheetTilesLen);
	
	// This is the memory we will manipulate to change the loaded tiles
	m_MapMemory = (u16*)BG_MAP_RAM(0);
}

// Load the tiles around a given x and y (Based on tile coordinates not pixels)
void CurrentMap::LoadTiles(int mapX, int mapY) 
{
	// Store map location so we can dervive the screen location to place the sprites
	m_MapX = mapX;
	m_MapY = mapY;
	
	// Since the memory of the tile memory is stored in 32x32 blocks we need to handle the memory in quadrants
	// Top left quadrant
	for(int i = 0; i < 32 * 32; ++i)
		m_MapMemory[i] = m_Tiles[(i % 32) + ((i / 32) * m_Width) + (mapX + (mapY * m_Width))];

	// Top right quadrant
	for(int i = 0; i < 32 * 32; ++i)
		m_MapMemory[1024+i] = m_Tiles[(i % 32) + ((i / 32) * m_Width) + 32 + (mapX + (mapY * m_Width))];

	// If this map is larger then 32 tiles high, we will load the additional tiles into memory
	if (m_Height > 32) {
		// Bottom left quadrant
		for(int i = 0; i < 32 * 32; ++i)
			m_MapMemory[2048+i] = m_Tiles[(i % 32) + (((i / 32) + 32) * m_Width) + (mapX + (mapY * m_Width))];

		// Bottom right quadrant
		for(int i = 0; i < 32 * 32; ++i)
			m_MapMemory[3072+i] = m_Tiles[(i % 32) + (((i / 32) + 32) * m_Width) + 32 + (mapX + (mapY * m_Width))];
	}
}

// Load the tiles around a given pixel x and y, will also set the screen offset to allow smooth transitions
void CurrentMap::FocusOn(int x, int y) 
{	
	/* These are the locations we'd like to place our screen offset
	   Start with the location we're focusing on (Gizmo's location)
	   Then load the pixel location of the top left loaded tile
	   Finally add an offset so Gizmo isn't directly centered		*/
	int screenX = x - m_MapX * 8 - 112;
	int screenY = y - m_MapY * 8 - 80;

	// Ensure we will not placing the offset outside the loaded tile array, so we place a top and bottom limit
	screenY = MathHelper::Clamp(screenY, 0,  m_Height * 8 - 192);

	// We will track if the tile map needs to be reloaded
	bool bReload = false;

	// Never let the loaded tiles or screen offset place the loaded tiles past the left edge of the map
	if (m_MapX <= 0 && screenX < 0)
	{
		screenX = 0;
		if (m_MapX < 0) {
			m_MapX = 0;
			bReload = true;
		}
	}

	// Never let the loaded tiles or screen offset place the loaded tiles past the right edge of the map
	if (m_MapX >= m_Width-64 && screenX > 256)
	{
		screenX = 256;
		if (m_MapX > m_Width-64) {
			m_MapX = m_Width-64;
			bReload = true;
		}
	}

	// If we're trying to place the offset outside of the loaded area were going to need to load a new set of tiles
	if (screenX < 0) {
		screenX += 128;
		m_MapX -=16;
		bReload = true;
	}
	if (screenX > 256) {
		screenX -= 128; 
		m_MapX += 16;
		bReload = true;
	}

	// Set the screen offset to the calculated and corrected location
	SetOffset(screenX, screenY);

	// If we need to load some new tiles then do that now
	if (bReload) {
		LoadTiles(m_MapX, m_MapY);
	}
}

// Sets the offset for the loaded tiles
void CurrentMap::SetOffset(int x, int y)
{
	// Registers set the offset on the tile background, we store it so we can calculate the screen location later
	REG_BG0HOFS = m_OffsetX = x;
	REG_BG0VOFS = m_OffsetY = y;
}

// Loads a given map from the maps in the pregenerated tilesheet.h header
void CurrentMap::LoadMap(int mapId) 
{
	// Check which map we're loading
	switch(mapId)
	{
		case 0:
			m_Width = tilesheet_Level1::Width;
			m_Height = tilesheet_Level1::Height;
			m_Tiles = (u16*)tilesheet_Level1::MapTiles;
			m_Bounds = (u8*)tilesheet_Level1::MapBounds;
			break;
		case 1:
			m_Width = tilesheet_Level2::Width;
			m_Height = tilesheet_Level2::Height;
			m_Tiles = (u16*)tilesheet_Level2::MapTiles;
			m_Bounds = (u8*)tilesheet_Level2::MapBounds;
			break;
		case 2:
			m_Width = tilesheet_Level3::Width;
			m_Height = tilesheet_Level3::Height;
			m_Tiles = (u16*)tilesheet_Level3::MapTiles;
			m_Bounds = (u8*)tilesheet_Level3::MapBounds;
			break;
		case 3:
			m_Width = tilesheet_Level4::Width;
			m_Height = tilesheet_Level4::Height;
			m_Tiles = (u16*)tilesheet_Level4::MapTiles;
			m_Bounds = (u8*)tilesheet_Level4::MapBounds;
			break;
	}

	// Load the tile sheet for this map
	LoadTiles(0,0);
	SetOffset(0,0);
}