#include "Gizmo.h"

// Constructors
Gizmo::Gizmo() : GameObject() { }
Gizmo::Gizmo(int x, int y, const Sprite sprite) : GameObject(x, y, sprite) { }

// Set the variables for gizmos death
void Gizmo::Died() 
{
	m_X = 64;
	m_Y = 96;
	m_DX = 0;
	m_DY = 0;
	m_Jumping = true;
	mmEffect( SFX_DEAD );
}

// Set the variables for gizmo jumping
void Gizmo::Jump() {
	m_DY = -52;
	m_Jumping = true;
}

// Update gizmo by reading input and doing collision detection
GizmoUpdateState Gizmo::Update(CurrentMap* map)
{
	// Since the scan keys is already called in the game we only need to poll keys that are currently being held here
	int keys = keysHeld();

	// This will allow a user to hit jump slightly before they land, makes the gameplay feel more fluid
	m_JumpAttempt--;
	if (keysDown() & KEY_A) 
		m_JumpAttempt = JUMP_FRAMES_WINDOW;

	// If the user recently attempted to jump we will have them jump and play the sound effect for it
	if ((m_JumpAttempt > 0) && (!m_Jumping)){
		Jump();
		mmEffect( SFX_LEAP );
		m_JumpAttempt = 0;
	}

	int speed = 1;

	// As long as both keys are not held
	if (!((keys & KEY_LEFT) && (keys & KEY_RIGHT))) {

		// If the user holds X we will double their speed
		if (keys & KEY_X)
			speed = 2;
		
		// If the user is not jumping they will stop moving
		if (!m_Jumping) 
			m_DX = 0;

		/* The user is capible of running at double speed and jumping at that speed, however we do not want
		   them to be able to change direction mid air and maintain that speed so we will check that here */

		// If the user wants to move left we will check if they are jumping and adjust their speed
		if (keys & KEY_LEFT) {
			if (!m_Jumping)
				m_DX = -speed;
			else 
				// Check if the user is changing direction, we don't want to slow them while still headed in the initial direction
				if (m_DX >= 0)
					m_DX = -1;

			// Ensure gizmo will face the correct direction
			m_Flipped = true;
		}
		
		// If the user wants to move left we will check if they are jumping and adjust their speed
		if (keys & KEY_RIGHT) {
			if (!m_Jumping)
				m_DX = speed;
			else 
				// Check if the user is changing direction, we don't want to slow them while still headed in the initial direction
				if (m_DX <= 0)
					m_DX = 1;

			// Ensure gizmo will face the correct direction
			m_Flipped = false;
		}
	}
	
	// Update Gizmo's horizontal location
	m_X = MathHelper::Clamp(m_X + m_DX, -5, map->GetWidth() * 8-27);


	// Ensure Gizmo does not move to quickly
	m_DY = MathHelper::Clamp(m_DY + 1, -60, 60);
	
	// Ensure if gizmo is falling, it is atleast at running speed, otherwise we will have issues on slopes
	if ((m_DY > 0) && (m_DY < 30))
		m_DY = 30;
	m_Y += m_DY / 15;

	// If Gizmo is inside the bounds of the map
	if (m_Y > 0) {

		// If gizmo is moving down, check blocks below him
		if (m_DY > 0)
			CollisionCheckDown(map, speed-1);
		// If gizmo moves right, check blocks to his right for collisions
		if (m_DX > 0)
			CollisionCheckRight(map);
		// if gizmo moves left, check blocks to his left for collisions
		if (m_DX < 0) 
			CollisionCheckLeft(map);
		// if gizmo moves up, check blocks above him for collisions
		if (m_DY < 0)
			CollisionCheckUp(map);

		// Always check the current block for diagonal bounds
		CollisionCheckDiagonal(map);

		// Check if gizmo has fallen off the map
		if (m_Y > map->GetHeight() * 8 - 33) 
		{
			Died();
			map->LoadTiles(0,0);
			return DiedUpdate;
		}
	}

	// Check if gizmo has reached the end of the current map
	if (m_X > map->GetWidth() * 8 - 33)
		return LevelClearUpdate;

	// Make sure the tiles are focused around gizmo's current position within the world
	map->FocusOn(m_X, m_Y);
	
	// If gizmo is falling we will consider him jumping
	if (m_DY > 0)
		m_Jumping = true;;

	// Check which animation we will play for gizmo
	if ((m_DX == 0) && (!m_Jumping))
		m_Sprite.UpdateSprite(Gizmo_IdleAnimation);
	else if (!m_Jumping)
	{
		m_Sprite.UpdateSprite(Gizmo_WalkAnimation);
	}
	else
		m_Sprite.UpdateSprite(Gizmo_JumpAnimation, (m_DY <-50) ? 0 : (m_DY < -40) ? 1 : (m_DY < 0) ? 2 : (m_DY < 45) ? 3 : 4);

	// Since no other returns happened, this was a normal update
	return NormalUpdate;
}

// Check if the block we're currently in is causing a diagonal collision
void Gizmo::CollisionCheckDiagonal(const CurrentMap* map) 
{
	// We're going to multi sample around the gizmo sprite
	int sampleX[2] = {(m_X + 13) / 32, (m_X + 17) / 32};
	int sampleY[2] = {(m_Y + 28) / 32, (m_Y + 32) / 32};

	// Grab the bounds array for the tiles
	u8* bounds = map->GetBounds();
	
	// The bounds array has 1/4th the number of entries
	int width = map->GetWidth() / 4;

	
	// For each of the sampled locations
	for (int a = 0; a < 2; ++a) 
		for (int b = 0; b < 2; ++b) {
			
			// If the sampled location exists within a diagonal block
			if (bounds[sampleX[a] + sampleY[b] * width] & BIT(7)) 
			{
				// Get gizmo's offset
				int offsetX = m_X + 16 - sampleX[a] * 32;
				int offsetY = m_Y + 32 - sampleY[b] * 32;
				
				// Check if the collision occured on a forward or backward facing slope
				if (bounds[sampleX[a] + sampleY[b] * width] & BIT(0)) 
				{
					// If Gizmo's offset place him inside the ground update his position
					if (offsetX < offsetY+1) 
					{
						m_Y = offsetX + sampleY[b] * 32 - 32;
						m_DY = 0;
						m_Jumping = false;
						return;
					}
				}
				else 
				{
					// If Gizmo's offset place him inside the ground update his position
					if (offsetX > (32 - offsetY-1)) 
					{
						m_Y = (32 - offsetX) + sampleY[b] * 32 - 32;
						m_DY = 0;
						m_Jumping = false;
						return;
					}
				}
			}
		}
}

// Check if the block we're walking into is causing a collision
void Gizmo::CollisionCheckLeft(const CurrentMap* map) 
{
	// We will multi sample along the y axis
	int boundX = (m_X + 4) / 32;
	int boundTopY =  (m_Y + 6)  / 32;
	int boundBottomY = (m_Y + 31 ) / 32;
	
	// Grab the bounds of the tiles
	u8* bounds = map->GetBounds();
	
	// The bounds array is 1/4th the size of the tile array
	int width = map->GetWidth() / 4;

	// If either location collides with a tile then we will adjust gizmo's location
	if (((~bounds[boundX + boundTopY * width] & BIT(7)) && (bounds[boundX + boundTopY * width] & BIT(3)))
		|| ((~bounds[boundX + boundBottomY * width] & BIT(7)) && (bounds[boundX + boundBottomY * width] & BIT(3))))
	{
		m_X = (boundX + 1) * 32 -4;
		m_DX = 0;
	}
}

// Check if the block we're walking into is causing a collision
void Gizmo::CollisionCheckRight(const CurrentMap* map)
{
	// We will multi sample along the y axis
	int boundX = (m_X + 28) / 32;
	int boundTopY =  (m_Y + 6)  / 32;
	int boundBottomY = (m_Y + 31 ) / 32;
	
	// Grab the bounds of the tiles
	u8* bounds = map->GetBounds();
	
	// The bounds array is 1/4th the size of the tile array
	int width = map->GetWidth() / 4;

	// If either location collides with a tile then we will adjust gizmo's location
	if (((~bounds[boundX + boundTopY * width] & BIT(7)) && (bounds[boundX + boundTopY * width] & BIT(1)))
		|| ((~bounds[boundX + boundBottomY * width] & BIT(7)) && (bounds[boundX + boundBottomY * width] & BIT(1))))
	{
		m_X = boundX * 32 - 28;
		m_DX = 0;
	}
}

// Check if the block we're jumping into is causing a collision
void Gizmo::CollisionCheckUp(const CurrentMap* map) 
{
	// We will multi sample along the x axis
	int boundLeftX = (m_X + 5) / 32;
	int boundRightX = (m_X + 26) / 32;
	int boundY = (m_Y + 6) / 32;
	
	// Grab the bounds of the tiles
	u8* bounds = map->GetBounds();
	
	// The bounds array is 1/4th the size of the tile array
	int width = map->GetWidth() / 4;

	// If either location collides with a tile then we will adjust gizmo's location
	if (((~bounds[boundLeftX + boundY * width] & BIT(7)) && (bounds[boundLeftX + boundY * width] & BIT(2)))
		|| ((~bounds[boundRightX + boundY * width] & BIT(7)) && (bounds[boundRightX + boundY * width] & BIT(2))))
	{
		m_Y = boundY * 32 + 26;
		m_DY = 0;
	}
}

// Check if the block we're on top of is causing a collision
void Gizmo::CollisionCheckDown(const CurrentMap* map, int speed)
{
	// We will multi sample along the x axis if the player is moving quickly we'll give them more leeway in collision detection
	int boundLeftX = (m_X + 4 - speed * 4) / 32;
	int boundRightX = (m_X + 27 + speed * 4) / 32;
	int boundY = (m_Y + 32) / 32;

	// Grab the bounds of the tiles
	u8* bounds = map->GetBounds();
	
	// The bounds array is 1/4th the size of the tile array
	int width = map->GetWidth() / 4;

	// Check if one of the tiles gizmo is over top of happens to be a diagonal tile
	if ((bounds[boundLeftX + boundY * width] & BIT(7)) || (bounds[boundRightX + boundY * width] & BIT(7)))
	{	
		// We will now only be concerned with the block gizmo is directly over top of
		// Make sure it is a flat tile then update his location (Diagonals are handled seperately
		boundLeftX = (m_X+16) / 32;
		if ((~bounds[boundLeftX + boundY * width] & BIT(7)) && (bounds[boundLeftX + boundY * width] & BIT(0)))
		{
			if (abs(m_Y + 32 - boundY*32) <= 4) {
				m_Y = boundY * 32 - 32;
				m_DY = 0;
				m_Jumping = false;
			}
		}
	}
	else 
	{
		// If either location collides with a tile then we will adjust gizmo's location
		if (((~bounds[boundLeftX + boundY * width] & BIT(7)) && (bounds[boundLeftX + boundY * width] & BIT(0)))
			|| ((~bounds[boundRightX + boundY * width] & BIT(7)) && (bounds[boundRightX + boundY * width] & BIT(0))))
		{
			if (abs(m_Y + 32 - boundY*32) <= 4) {
				m_Y = boundY * 32 - 32;
				m_DY = 0;
				m_Jumping = false;
			}
		}
	}
}
