#include "Snail.h"

// Constructors
Snail::Snail() : GameObject(), m_Dying(false) { }
Snail::Snail(int x, int y, Sprite sprite) : GameObject(x, y, sprite), m_Dying(false) { }

// Updates the animations, aswell as tells the pathfinding algorithm to run, returns true when the snail is dead
bool Snail::Update(const CurrentMap* map)
{
	// If the snail is not dying we will access the pathfinding
	if (!m_Dying) {
		switch(m_Sprite.GetCurrentState()) 
		{
		case BlueSnail_Walk:
			Walk(map);
			break;
		case RedSnail_Walk:
			Walk(map);
			break;
		default:
			break;
		}
	}
	else
	{
		// The snail is dying, we need to check when the last part of the animation has displayed then remove the snail from memory
		if (m_Sprite.GetAnimationFrame() >= 39){
			Clean();
			return true;
		}
	}
	
	// Update the sprite's animation
	m_Sprite.UpdateSprite();
	return false;
}


// Checks if the snail has collided with gizmo and what type of collision it was
SnailCollisionState Snail::CollisionCheck(Gizmo* gizmo, CurrentMap* map)
{
	// If the snail is not dying we will check for collisions
	if (!m_Dying)
	{
		// Find the difference in location between the snail and gizmo
		int OffsetX = abs(m_X - gizmo->GetX());
		int OffsetY = m_Y - gizmo->GetY();

		// If the distances are such that a collision has occured we will check which type of collision
		if ((OffsetX < 14) && ((OffsetY < 14) && (OffsetY >= 0)))
		{
			// If the Y difference is greater then 8 then gizmo hit the top of the snail and it will die
			if (OffsetY >= 8) {
				m_Sprite.UpdateSprite((m_Sprite.GetCurrentState() == BlueSnail_Walk) ? BlueSnail_Death : RedSnail_Death);
				m_Dying = true;
				gizmo->Jump();
				return SnailDiedCollision;
			}

			// Otherwise gizmo got hit by the snail and will die
			else {
				gizmo->Died();	
				map->LoadTiles(0,0);
				return GizmoDiedCollision;
			}
		}
	}

	// In the case of no collision we will inform the game to take no action
	return NoCollision;
}

// Path finding algorithm to find the direction the snail should walk
void Snail::Walk(const CurrentMap* map)
{
	// Get the location the snail is walking over
	int mapX = (m_X + ((m_Flipped) ? 32 : 0)) / 32;
	int mapY = (m_Y + 32) / 32;
	
	// Grab the tile bounds array
	u8* bounds = map->GetBounds();

	// The bounds array is 1/4th the size of the tile array
	int width = map->GetWidth() / 4;

	// Make sure the snail bounces off map bounds
	bool MapBound = false;
	if (m_X <= -5)
	{	
		m_Flipped = true;
		MapBound = true;
	}
	if (m_X >= width*32-32)
	{
		m_Flipped = false;
		MapBound = true;
	}
	
	// Check if the snail is about to walk over an edge, if it is turn around
	if ((!MapBound) && (!((~bounds[mapX + mapY * width] & BIT(7)) && (bounds[mapX + mapY * width] & BIT(0)))))
	{
		m_Flipped = !m_Flipped;
	}

	// Move in the direction the snail is facing
	m_X += (m_Flipped) ? 1 : -1;

}